modules = [

    #"logic",

    "decode",
    "decoder",
    "encoder",
    "arbiter",

    'halfadder',
    'fulladder',
    "arith",

    "compare",

    "barrel",

    "register",

    "siso",
    "sipo",
    "piso",
    "pipo",

    "ring",
    "cascadedring",

    "johnson",

    "counter",
    "countermod",
    "counterload",

    "RAM",
    "ROM",

    "primitives",
    "operator",

    "reg_file",
]



import math
import magma as m
from mantle import And, DefineAdd
from mantle import Mux
from mantle import Add
from mantle import Register
from mantle import Decode


__all__  = ['DefineCounter', 'Counter']
__all__ += ['DefineUpCounter', 'UpCounter']
__all__ += ['DefineDownCounter', 'DownCounter']
__all__ += ['DefineUpDownCounter', 'UpDownCounter']


def counter_name(name, incr, has_ce, has_reset, cin, cout):
    if incr != 1: name += f'_{incr}'
    if has_ce: name += 'CE'
    if has_reset:  name += 'R'
    if cin: name += '_CIN'
    if cout: name += '_COUT'
    return name


def DefineCounter(n, cin=False, cout=True, incr=1,
                  has_ce=False, has_reset=False):
    """
    Create an n-bit counter with a given increment.

    O : m.Out(m.UInt(n)), COUT : m.Out(m.Bit)
    """
    name_ = counter_name(f'Counter{n}', incr, has_ce, has_reset, cin, cout)
    args = {}
    if cin:
        args["CIN"] = m.In(m.Bit)
    args["O"] = m.Out(m.UInt[n])
    if cout:
        args["COUT"] = m.Out(m.Bit)

    class _Counter(m.Circuit):
        name = name_
        io = m.IO(**args)
        io += m.ClockIO(has_ce, has_reset)

        add = DefineAdd(n, cin=cin, cout=cout)()
        reg = Register(n, has_ce=has_ce, has_reset=has_reset)

        m.wire(reg.O, add.I0)
        m.wire(m.array(incr, n), add.I1)
        reg(add.O)

        next = False
        if next:
            m.wire(add.O, io.O)
        else:
            m.wire(reg.O, io.O)
        if cin:
            m.wire(io.CIN, add.CIN)
        if cout:
            m.wire(add.COUT, io.COUT)

    return _Counter


def Counter(n, cin=False, cout=True, incr=1,
            has_ce=False, has_reset=False, **kwargs):
    """Construct a n-bit up counter"""
    defn = DefineUpCounter(n, cin=cin, cout=cout, incr=incr, has_ce=has_ce,
                           has_reset=has_reset)
    return defn(**kwargs)


DefineUpCounter = DefineCounter
UpCounter = Counter


def DefineDownCounter(n, cin=False, cout=True, decr=1,
                      has_ce=False, has_reset=False):
    """
    Create an n-bit down counter.

    O : m.Out(m.UInt(n)), COUT : m.Out(m.Bit)
    """
    incr = (1 << n) - (decr)
    return DefineCounter(n, cin=cin, cout=cout, incr=incr, has_ce=has_ce,
                         has_reset=has_reset)


def DownCounter(n, cin=False, cout=True, decr=1,
                has_ce=False, has_reset=False, **kwargs):
    defn = DefineDownCounter(n, cin=cin, cout=cout, decr=decr,
                             has_ce=has_ce, has_reset=has_reset)
    return defn(**kwargs)


def DefineUpDownCounter(n, cout=True, has_ce=False, has_reset=False):
    """
    Create an n-bit up-down counter.

    U : In(Bit), D : In(Bit), O : Out(UInt(n)), COUT : Out(Bit)
    """
    name_ = counter_name(f'UpDownCounter{n}', 1, has_ce,
                         has_reset, False, cout)

    args = {"U": m.In(m.Bit), "D": m.In(m.Bit), "O": m.Out(m.UInt[n])}
    if cout:
        args["COUT"] = m.Out(m.Bit)

    class _Counter(m.Circuit):
        name = name_
        io = m.IO(**args)
        io += m.ClockIO(has_ce, has_reset)

        add = Add(n, cin=True, cout=cout)
        reg = Register(n, has_ce=has_ce, has_reset=has_reset)
    
        m.wire(reg.O, add.I0)
        m.wire(m.array(n * [io.D]), add.I1)
        m.wire(io.U, add.CIN)
    
        reg(add)
    
        next = False
        if next:
            m.wire(add.O, io.O)
        else:
            m.wire(reg.O, io.O)
        if cout:
            m.wire(add.COUT, io.COUT)
    
    return _Counter


def UpDownCounter(n, cout=True, has_ce=False, has_reset=False, **kwargs):
    """Construct an n-bit updown counter."""
    defn = DefineUpDownCounter(n, cout=cout,
                               has_ce=has_ce, has_reset=has_reset)
    return defn(**kwargs)


def DefineCeilFloorUpDownCounter(n, has_ce=False, has_reset=False):
    """
    Counter between 0 and n - 1 that uses the minimum number of bits.

    This counter counts up by 1 if U is True and D is False, down by 1 if U is
    False and D is True, no change if both are False or True.

    Also, unlike normal updown counter, up by 1 has no effect if already at (n -
    1) and down by 1 has nefect if already at 0.

    :param n: The value the counter counts up to
    :param has_ce: Whether this counter should a clock-enable input
    :return: An updown counter circuit
    """
    num_bits = math.ceil(math.log(n, 2))

    class _Counter(m.Circuit):
        name = (f"CeilFloorUpDownCounter_m{str(n)}_hasCE{str(has_ce)}_"
                f"hasReset{str(has_reset)}")
        io = m.IO(U=m.In(m.Bit), D=m.In(m.Bit), O=m.Out(m.UInt[num_bits]))
        io += m.ClockIO(has_ce, has_reset)

        up_down_counter = UpDownCounter(num_bits, False, has_ce, has_reset)
        at_max = Decode(n - 1, up_down_counter.O.N)(up_down_counter.O)
        at_min = Decode(0, up_down_counter.O.N)(up_down_counter.O)

        # Only increase counter if input says to increase and not at max.
        should_increase = ~at_max & io.U & ~io.D
        m.wire(should_increase, up_down_counter.U)
        # Only decrease counter if input says to decrease and not at min.
        should_decrease = ~at_min & ~io.U & io.D
        m.wire(should_decrease, up_down_counter.D)

        m.wire(up_down_counter.O, io.O)

    return _Counter


def CeilFloorUpDownCounter(m, has_ce=False, has_reset=False):
    return DefineCeilFloorUpDownCounter(m, has_ce, has_reset)()



import math
from magma import *
import coreir
from magma.bitutils import clog2, seq2int, int2seq, fun2seq
from types import FunctionType
from collections.abc import Sequence
from hwtypes import BitVector


@cache_definition
def DeclareCoreirLUT(N, init):

    def _simulate(self, value_store, state_store):
        in_ = value_store.get_value(getattr(self, "in"))
        value_store.set_value(self.out, [bool(i) for i in int2seq(init, 2 ** N)][seq2int(in_)])

    args = {'in': In(Bits[N]), 'out': Out(Bit)}

    class _coreir_lut(Circuit):
        name = "coreir_lut{}".format(N)
        io = IO(**args)
        simulate = _simulate
        coreir_name = "lutN"
        coreir_lib  = "commonlib"
        coreir_genargs = {"N": N}
        coreir_configargs = {"init": BitVector[1 << N](init)}

    return _coreir_lut

def DefineLUT(init, N):
    io = []
    for i in range(N):
        io += ["I{}".format(i), In(Bit)]
    io += ["O", Out(Bit)]

    class LUT(Circuit):
        name = "LUT{}_{}".format(N, init)
        IO = io
        @classmethod
        def definition(cls):
            lutN = DeclareCoreirLUT(N, init)()
            for i in range(N):
                wire(getattr(lutN, "in")[i], getattr(cls, "I{}".format(i)))
            wire(lutN.out, cls.O)

    return LUT

def LUT(init, N=None, **kwargs):
    """
    n-bit LUT

    I0 : In(Bit), I1 : In(Bit), ..., In : In(Bit),  O : Out(Bit)
    """

    if isinstance(init, FunctionType):
        init = fun2seq(init, 1<<N)

    if isinstance(init, Sequence):
        if N is not None:
            if 2 ** N < len(init):
                raise ValueError("init is too large for N={}".format(N))
        else:
            N = clog2(len(init))
        init = seq2int(init)
    else:
        if N is None:
            raise ValueError("N requires for not sequence init")

    return DefineLUT(init, N)()



from hwtypes import BitVector
from .logic import (
    And      , DefineAnd   , ReduceAnd ,
    NAnd     , DefineNAnd  , ReduceNAnd,
    Or       , DefineOr    , ReduceOr  ,
    NOr      , DefineNOr   , ReduceNOr ,
    XOr      , DefineXOr   , ReduceXOr ,
    NXOr     , DefineNXOr  , ReduceNXOr,
    Invert   , DefineInvert,
    Wire   , DefineWire,
    Not      ,
    DefineStaticLeftShift  , StaticLeftShift  , static_left_shift  ,
    DefineStaticRightShift , StaticRightShift , static_right_shift ,
    DefineLSL, LSL ,
    DefineLSR, LSR,
)

#from .fulladder import FullAdder
#from .halfadder import HalfAdder
from .arith import DefineAdd, DefineSub, DefineNegate, DefineASR, ASR, \
    DefineCoreirAdd, DefineCoreirMul, DefineMul, DefineUDiv, DefineSDiv, \
    DefineUMod, DefineSMod

from .FF import FF, DFF, DefineDFF, DefineCoreirReg
from .LUT import LUT

from .MUX import Mux2, Mux4, Mux8, Mux16, DefineMux, Mux, DefineCoreirMux
from .memory import DefineRAM, DefineMemory, DefineCoreirMem
from .compare import \
    DefineEQ, DefineCoreirEq, EQ, \
    DefineNE, DefineCoreirNeq, NE, \
    DefineULT, DefineCoreirUlt, ULT, \
    DefineULE, DefineCoreirUle, ULE, \
    DefineUGT, DefineCoreirUgt, UGT, \
    DefineUGE, DefineCoreirUge, UGE, \
    DefineSLT, DefineCoreirSlt, SLT, \
    DefineSLE, DefineCoreirSle, SLE, \
    DefineSGT, DefineCoreirSgt, SGT, \
    DefineSGE, DefineCoreirSge, SGE


from magma import bits, cache_definition, Circuit, Bits, wire, Out, In, Bit
import magma as m
from .util import DeclareCoreirCircuit

def DefineCoreirConst(width, value):
    def simulate_coreir_const(self, value_store, state_store):
        value_store.set_value(self.O, value)
    return DeclareCoreirCircuit(f"coreir_const{width}{value}", "O", Out(Bits[ width ]),
            coreir_name="const", coreir_lib="coreir",
            coreir_genargs={"width": width},
            coreir_configargs={"value": BitVector[width](value)},
            simulate=simulate_coreir_const)

def DefineCorebitConst(value):
    if not isinstance(value, bool) and not (isinstance(value, int) and value in {0, 1}):
        raise ValueError("DefineCorebitConst expects a boolean value or 0 or 1")
    def simulate_corebit_const(self, value_store, state_store):
        value_store.set_value(self.O, value)
    return DeclareCoreirCircuit(f"corebit_const{value}", "O", Out(Bit),
            coreir_name="const", coreir_lib="corebit",
            default_kwargs={"value": bool(value)},
            simulate=simulate_corebit_const)


def DefineTerm(width):
    def simulate_term(self, value_store, state_store):
        pass
    return DeclareCoreirCircuit(f"term", "I", In(Bits[ width ]),
                                coreir_name="term", coreir_lib="coreir",
                                coreir_genargs={"width": width},
                                simulate=simulate_term)


def Term(width):
    return DefineTerm(width)()


def DefineCorebitTerm():
    def simulate_corebit_term(self, value_store, state_store):
        pass
    return DeclareCoreirCircuit(f"corebit_term", "I", In(Bit),
                                coreir_name="term", coreir_lib="corebit",
                                simulate=simulate_corebit_term)


def CorebitTerm():
    return DefineCorebitTerm()()

from .register import DefineRegister, Register, register



from magma import *
import magma as m


@cache_definition
def DeclareAdd(N, cin=False, cout=False):
    has_cin = cin
    has_cout = cout
    T = Bits[N]
    IO_ = ['I0', In(T), 'I1', In(T), 'O', Out(T)]
    name_ = "Add{}".format(N)
    if has_cout:
        IO_ += ['COUT', Out(Bit)]
        name_ += "_cout"
    if has_cin:
        IO_ += ['CIN', In(Bit)]
        name_ += "_cin"
    class Add(Circuit):
        # Underscores because there's some weird scoping issue here with Python
        # when trying to capture name and IO
        name = name_
        IO = IO_
    return Add


def add(*args, **kwargs):
    width = get_length(args[0])
    if not all(get_length(arg) == width for arg in args):
        # TODO: Something more specific than a ValueError?
        raise ValueError("Arguments to add should all be the same width")
    if not all(isinstance(arg, BitsType) for arg in args):
        # TODO: Something more specific than a ValueError?
        raise ValueError("Arguments to add should be all Bits"
                " {}".format([(arg, type(arg)) for arg in args]))
    adders = [Add(width, **kwargs) for _ in range(len(args) - 1)]
    curr = adders[0]
    wire(args[0], curr.I0)
    wire(args[1], curr.I1)
    if len(args) > 2:
        next_ = adders[1]
        for i in range(1, len(adders)):
            next_ = adders[i]
            wire(curr.O, next_.I0)
            wire(args[i + 1], next_.I1)
            curr = next_
    return curr.O


@cache_definition
def DeclareSub(N, cin=False, cout=False, T=m.Bits):
    has_cin = cin
    has_cout = cout
    T = T[N]
    IO_ = ['I0', In(T), 'I1', In(T), 'O', Out(T)]
    name_ = "Sub{}".format(N)
    if has_cout:
        IO_ += ['COUT', Out(Bit)]
        name_ += "_cout"
    if has_cin:
        IO_ += ['CIN', In(Bit)]
        name_ += "_cin"
    class Sub(Circuit):
        # Underscores because there's some weird scoping issue here with Python
        # when trying to capture name and IO
        name = name_
        IO = IO_
    return Sub


def sub(*args, **kwargs):
    width = get_length(args[0])
    if not all(get_length(arg) == width for arg in args):
        # TODO: Something more specific than a ValueError?
        raise ValueError("Arguments to sub should all be the same width")
    if not all(isinstance(arg, BitsType) for arg in args):
        # TODO: Something more specific than a ValueError?
        raise ValueError("Arguments to sub should be all Bits"
                " {}".format([(arg, type(arg)) for arg in args]))
    subbers = [Sub(width, **kwargs) for _ in range(len(args) - 1)]
    curr = subbers[0]
    wire(args[0], curr.I0)
    wire(args[1], curr.I1)
    if len(args) > 2:
        next_ = subbers[1]
        for i in range(1, len(subbers)):
            next_ = subbers[i]
            wire(curr.O, next_.I0)
            wire(args[i + 1], next_.I1)
            curr = next_
    return curr.O

def DeclareNegate(width):
    T = Bits[width]
    class _Negate(Circuit):
        name = 'Negate{}'.format(width)
        IO = ['I', In(T), 'O', Out(T)]
    return _Negate


def DeclareASR(width):
    T = Bits[width]
    class _ASR(Circuit):
        name = 'ASR{}'.format(width)
        IO = ["I0", In(T), "I1", In(T), "O", Out(T)]
    return _ASR



from magma import *
import magma as m
from hwtypes import BitVector
import hwtypes as ht
from magma.compatibility import IntegerTypes
import operator
from functools import reduce
from .util import DeclareCoreirCircuit


def get_length(value):
    if isinstance(value, m.Digital):
        return None
    if isinstance(value, m.Array):
        return len(value)
    raise NotImplementedError(f"Cannot get_length of {type(value)}")


def _make_lambda(op):
    return lambda x, y: op()(x, y)


def _get_inputs(io, height):
    return [getattr(io, f"I{i}") for i in range(height)]


def DefineFoldOp(DefineOp, name, height, width):
    T = Bit if width is None else Bits[width]
    args = {f"I{i}": In(T) for i in range(height)}
    args.update({"O": Out(T)})
    name_ = name

    class _FoldOp(m.Circuit):
        name = f"fold_{name_}{height}{width}"
        io = m.IO(**args)

        # NOTE(rsetaluri): Because of scoping rules in the class definition, we
        # can not use 'io' neither in list comprehensions nor as a captured
        # variable in a lambda. Therfore we use simple iteration-based list
        # construction and delegate lambda construction to a helper function.
        reduce_args = []
        for i in range(height):
            reduce_args.append(getattr(io, f"I{i}"))
        Op2 = DefineOp(2, width)
        wire(reduce(_make_lambda(Op2), reduce_args), io.O)

    return _FoldOp


def declare_bit_binop(name, python_op):
    def simulate(self, value_store, state_store):
        I0 = ht.Bit(value_store.get_value(self.I0))
        I1 = ht.Bit(value_store.get_value(self.I1))
        #O = python_op(I0, I1).as_bool_list()[0]
        O = int(python_op(I0, I1))
        value_store.set_value(self.O, O)

    return DeclareCoreirCircuit("{}".format(name),
                                'I0', In(Bit), 'I1', In(Bit), 'O', Out(Bit),
                                simulate=simulate,
                                verilog_name = "coreir_" + name,
                                firrtl_op  = name,
                                coreir_lib = "corebit")


def DefineCoreirReduce(op_name, python_op, width):
    def simulate(self, value_store, state_store):
        in_ = BitVector[width](value_store.get_value(self.I))
        O = reduce(python_op, in_)
        value_store.set_value(self.O, O)
    decl = DeclareCoreirCircuit(op_name, "I", In(Bits[width]), "O", Out(Bit),
            coreir_name = op_name,
            coreir_lib = "coreir",
            coreir_genargs = {"width": width},
            simulate = simulate)
    return decl

def DefineCoreirReduceAnd(width):
    return DefineCoreirReduce("andr", operator.and_, width)

def DefineCoreirReduceOr(width):
    return DefineCoreirReduce("orr", operator.or_, width)

def DefineCoreirReduceXOr(width):
    return DefineCoreirReduce("xorr", operator.xor, width)
    # class ReduceAnd(Circuit):
    #     name = f"reduce_and_{width}"
    #     IO = ["I", In(Bits(width)), "O", Out(Bit)]
    #     @classmethod
    #     def definition(io):
    #         circ = decl()
    #         wire(getattr(circ, "in"), io.I)
    #         wire(circ.out, io.O)
    # return ReduceAnd


def declare_bits_binop(name, python_op):

    def Declare(N):
        def simulate(self, value_store, state_store):
            I0 = BitVector[N](value_store.get_value(self.I0))
            I1 = BitVector[N](value_store.get_value(self.I1))
            O = python_op(I0, I1).as_bool_list()
            # print(f"{python_op}({I0}, {I1}), {out}")
            value_store.set_value(self.O, O)
        T = Bits[N]
        return DeclareCoreirCircuit("{}{}".format(name, N),
                                    'I0', In(T), 'I1', In(T), 'O', Out(T),
                                    simulate       = simulate,
                                    verilog_name   = "coreir_" + name,
                                    coreir_name    = name,
                                    coreir_lib     = "coreir",
                                    coreir_genargs = {"width": N})

    return Declare


@cache_definition
def DefineOp(op_name, DefineCoreirReduce, height, width):
    T = Bit if width is None else Bits[width]
    args = {f"I{i}": In(T) for i in range(height)}
    args.update({"O": Out(T)})

    class _Op(m.Circuit):
        name = f"{op_name}{height}x{width}"
        io = m.IO(**args)

        if width is None:
            reduce_ = DefineCoreirReduce(height)()
            for j in range(height):
                wire(reduce_.I[j], getattr(io, f"I{j}"))
            wire(reduce_.O, io.O)
        else:
            for i in range(width):
                reduce_ = DefineCoreirReduce(height)()
                for j in range(height):
                    wire(reduce_.I[j], getattr(io, f"I{j}")[i])
                wire(reduce_.O, io.O[i])

    return _Op

def DefineAnd(height=2, width=None):
    if height is 2:
        if width is None:
            return declare_bit_binop("and", operator.and_)
        else:
            return declare_bits_binop("and", operator.and_)(width)
    return DefineOp("And", DefineCoreirReduceAnd, height, width)


def And(height, width=None, **kwargs):
    return DefineAnd(height, width)(**kwargs)

def ReduceAnd(height=2, **kwargs):
    return uncurry(And(height, **kwargs))


def DefineNAnd(height=2, width=None):
    if width is None:
        T = Bit
    else:
        T = Bits[width]
    decl = []
    for i in range(height):
        decl += ["I{}".format(i), In(T)]
    decl += ["O", Out(T)]

    class _NAnd(Circuit):
        name = f"NAnd{height}{width}"
        io = IO(**dict(zip(decl[::2], decl[1::2])))
        inputs = _get_inputs(io, height)
        if width is None:
            inv = Not()
        else:
            inv = Invert(width)
        O = inv(And(height, width)(*inputs))
        wire(O, io.O)

    return _NAnd


def NAnd(height, width=None, **kwargs):
    return DefineNAnd(height, width)(**kwargs)

def ReduceNAnd(height=2, **kwargs):
    return uncurry(NAnd(height, **kwargs))


def simulate_bit_not(self, value_store, state_store):
    _in = ht.Bit(value_store.get_value(self.I))
    #O = (~_in).as_bool_list()[0]
    O = ~_in
    value_store.set_value(self.O, bool(O))


Not = DeclareCoreirCircuit("not", 'I', In(Bit), 'O', Out(Bit),
    simulate=simulate_bit_not, verilog_name="coreir_bitnot", coreir_lib="corebit")


def not_(arg, **kwargs):
    return Not(get_length(arg), **kwargs)(arg)


def DefineOr(height=2, width=None):
    if height is 2:
        if width is None:
            return declare_bit_binop("or", operator.or_)
        else:
            return declare_bits_binop("or", operator.or_)(width)
    return DefineOp("Or", DefineCoreirReduceOr, height, width)


def Or(height, width=None, **kwargs):
    return DefineOr(height, width)(**kwargs)

def ReduceOr(height=2, **kwargs):
    return uncurry(Or(height, **kwargs))


def DefineNOr(height=2, width=None):
    if width is None:
        T = Bit
    else:
        T = Bits[width]
    decl = []
    for i in range(height):
        decl += ["I{}".format(i), In(T)]
    decl += ["O", Out(T)]

    class _NOr(Circuit):
        name = f"NOr{height}{width}"
        io = IO(**dict(zip(decl[::2], decl[1::2])))
        inputs = _get_inputs(io, height)
        if width is None:
            inv = Not()
        else:
            inv = Invert(width)
        O = inv(Or(height, width)(*inputs))
        wire(O, io.O)

    return _NOr


def NOr(height, width=None, **kwargs):
    return DefineNOr(height, width)(**kwargs)

def ReduceNOr(height=2, **kwargs):
    return uncurry(NOr(height, **kwargs))


def DefineXOr(height=2, width=None):
    if height is 2:
        if width is None:
            return declare_bit_binop("xor", operator.xor)
        else:
            return declare_bits_binop("xor", operator.xor)(width)
    else:
        return DefineFoldOp(DefineXOr, "xor", height, width)


def XOr(height, width=None, **kwargs):
    return DefineXOr(height, width)(**kwargs)

def ReduceXOr(height=2, **kwargs):
    return uncurry(XOr(height, **kwargs))


def DefineNXOr(height=2, width=None):
    if width is None:
        T = Bit
    else:
        T = Bits[width]
    decl = []
    for i in range(height):
        decl += ["I{}".format(i), In(T)]
    decl += ["O", Out(T)]

    class _NXOr(Circuit):
        name = f"NXOr{height}{width}"
        io = IO(**dict(zip(decl[::2], decl[1::2])))
        inputs = _get_inputs(io, height)
        if width is None:
            inv = Not()
        else:
            inv = Invert(width)
        O = inv(XOr(height, width)(*inputs))
        wire(O, io.O)

    return _NXOr


def NXOr(height, width=None, **kwargs):
    return DefineNXOr(height, width)(**kwargs)

def ReduceNXOr(height=2, **kwargs):
    return uncurry(NXOr(height, **kwargs))

def DefineInvert(width):
    T = Bits[width]


    def simulate_bits_invert(self, value_store, state_store):
        _in = BitVector[width](value_store.get_value(self.I))
        O = (~_in).as_bool_list()
        value_store.set_value(self.O, O)
    return DeclareCoreirCircuit("Invert{}".format(width),
            'I', In(T), 'O', Out(T),
            simulate       = simulate_bits_invert,
            verilog_name   = "coreir_not",
            coreir_name    = "not",
            coreir_lib     = "coreir",
            coreir_genargs = {"width": width})

def Invert(width=None, **kwargs):
    return DefineInvert(width)(**kwargs)


def simulate_wire(self, value_store, state_store):
    value_store.set_value(self.O, value_store.get_value(self.I))


def DefineWire(width):
    if width is None:
        T = Bit
        coreir_lib = "corebit"
        genargs = None
    else:
        T = Bits[width]
        coreir_lib = "coreir"
        genargs = {"width": width}
    return DeclareCoreirCircuit("Wire{}".format(width),
            'I', In(T), 'O', Out(T),
            simulate       = simulate_wire,
            coreir_name    = "wire",
            coreir_lib     = coreir_lib,
            coreir_genargs = genargs)

def Wire(width=None, **kwargs):
    return DefineWire(width)(**kwargs)


def invert(arg, **kwargs):
    return Invert(get_length(arg), **kwargs)(arg)


def DefineLSL(width):
    T = Bits[width]
    def simulate(self, value_store, state_store):
        I0 = BitVector[width](value_store.get_value(self.I0))
        I1 = BitVector[width](value_store.get_value(self.I1))
        O = (I0 << I1).as_bool_list()
        value_store.set_value(self.O, O)

    return DeclareCoreirCircuit("shl{}".format(width), 'I0', In(T), 'I1',
            In(UInt[width]), 'O', Out(T), verilog_name="coreir_shl",
            coreir_name="shl", coreir_lib="coreir", simulate=simulate,
            coreir_genargs={"width": width})


def LSL(width, **kwargs):
    return DefineLSL(width)(**kwargs)


def DefineLSR(width):
    T = Bits[width]
    def simulate(self, value_store, state_store):
        I0 = BitVector[width](value_store.get_value(self.I0))
        I1 = BitVector[width](value_store.get_value(self.I1))
        O = (I0 << I1).as_bool_list()
        value_store.set_value(self.O, O)

    return DeclareCoreirCircuit("lshr{}".format(width), 'I0', In(T), 'I1',
            In(UInt[width]), 'O', Out(T), verilog_name="coreir_lshr",
            coreir_name="lshr", coreir_lib="coreir", simulate=simulate,
            coreir_genargs={"width": width})


def LSR(width, **kwargs):
    return DefineLSR(width)(**kwargs)


def DefineStaticLeftShift(width, shift_amount):
    T = Bits[width]
    class _StaticLeftShift(Circuit):
        name = 'StaticLeftShift_{}{}'.format(width, shift_amount)

        IO = ["I", In(T), "O", Out(T)]

        @classmethod
        def definition(io):
            output = concat(io.I[shift_amount:width], bits(0, shift_amount))
            wire(output, io.O)
    return _StaticLeftShift

def StaticLeftShift(width, shift_amount, **kwargs):
    return DefineStaticLeftShift(width, shift_amount)(**kwargs)

def static_left_shift(arg, shift_amount, **kwargs):
    width = get_length(arg)
    return StaticLeftShift(width, shift_amount, **kwargs)(arg)

def DefineStaticRightShift(width, shift_amount):
    T = Bits[width]
    class _StaticRightShift(Circuit):
        name = 'StaticRightShift_{}{}'.format(width, shift_amount)

        IO = ["I", In(T), "O", Out(T)]

        @classmethod
        def definition(io):
            output = concat(bits(0, shift_amount), io.I[:width-shift_amount])
            wire(output, io.O)
    return _StaticRightShift

def StaticRightShift(width, shift_amount, **kwargs):
    return DefineStaticRightShift(width, shift_amount)(**kwargs)

def static_right_shift(arg, shift_amount, **kwargs):
    width = get_length(arg)
    return StaticRightShift(width, shift_amount, **kwargs)(arg)



import operator
from hwtypes import BitVector
import magma as m
from .arith import declare_binop, get_length
from .logic import not_, XOr
from .util import DeclareCoreirCircuit


def DefineEQ(width):

    def _simulate(self, value_store, state_store):
        I0 = BitVector[width](value_store.get_value(self.I0))
        I1 = BitVector[width](value_store.get_value(self.I1))
        O = int(operator.eq(I0, I1))
        value_store.set_value(self.O, O)

    if width is None:

        class corebit_eq(m.Circuit):
            name = "corebit_eq"
            io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), O=m.Out(m.Bit))
            m.wire(io.O, not_(XOr(2, None)(io.I0, io.I1)))

        return _corebit_eq

    T = m.Bits[width]
    return DeclareCoreirCircuit("coreir_eq_{}".format(width),
                                'I0', m.In(T), 'I1', m.In(T),
                                'O', m.Out(m.Bit),
                                stateful=False,
                                simulate=_simulate,
                                verilog_name="coreir_eq",
                                coreir_name="eq",
                                coreir_lib="coreir",
                                coreir_genargs={"width": width})


DefineCoreirEq = DefineEQ


def EQ(n, **kwargs):
    return DefineEQ(n)(**kwargs)


DefineCoreirNeq = declare_binop("neq", operator.ne, out_type=m.Bit)


def DefineNE(n):
    T = m.Bit if n is None else m.Bits[n]

    class _NE(m.Circuit):
        name = f"NE{n}"
        io = m.IO(I0=m.In(T), I1=m.In(T), O=m.Out(m.Bit))
        eq = EQ(n)
        O = not_(eq(io.I0, io.I1))
        m.wire(O, io.O)

    return _NE


def NE(n, **kwargs):
    return DefineNE(n)(**kwargs)


DefineCoreirUlt = declare_binop("ult", operator.lt, out_type=m.Bit)


def DefineULT(n):
    return DefineCoreirUlt(n)


def ULT(n, **kwargs):
    return DefineULT(n)(**kwargs)


DefineCoreirUle = declare_binop("ule", operator.le, out_type=m.Bit)


def DefineULE(n):
    return DefineCoreirUle(n)


def ULE(n, **kwargs):
    return DefineULE(n)(**kwargs)


DefineCoreirUgt = declare_binop("ugt", operator.gt, out_type=m.Bit)


def DefineUGT(n):
    return DefineCoreirUgt(n)


def UGT(n, **kwargs):
    return DefineUGT(n)(**kwargs)


DefineCoreirUge = declare_binop("uge", operator.ge, out_type=m.Bit)


def DefineUGE(n):
    return DefineCoreirUge(n)


def UGE(n, **kwargs):
    return DefineUGE(n)(**kwargs)


DefineCoreirSlt = declare_binop("slt", operator.lt, out_type=m.Bit)


def DefineSLT(n):
    return DefineCoreirSlt(n)


def SLT(n, **kwargs):
    return DefineSLT(n)(**kwargs)


DefineCoreirSle = declare_binop("sle", operator.le, out_type=m.Bit)


def DefineSLE(n):
    return DefineCoreirSle(n)


def SLE(n, **kwargs):
    return DefineSLE(n)(**kwargs)


DefineCoreirSgt = declare_binop("sgt", operator.gt, out_type=m.Bit)


def DefineSGT(n):
    return DefineCoreirSgt(n)


def SGT(n, **kwargs):
    return DefineSGT(n)(**kwargs)


DefineCoreirSge = declare_binop("sge", operator.ge, out_type=m.Bit)


def DefineSGE(n):
    return DefineCoreirSge(n)


def SGE(n, **kwargs):
    return DefineSGE(n)(**kwargs)



import math
from magma import *
from mantle import And, Decode, Or
from .counter import Counter, counter_name


__all__ = ['DefineUpCounterModM', 'UpCounterModM']
__all__ += ['DefineCounterModM', 'CounterModM', 'SizedCounterModM']


def DefineCounterModM(m, n, cin=False, cout=True, incr=1,
                      has_ce=False, has_reset=False):
    """Create an n-bit mod-m counter"""

    name_ = counter_name(f'Counter{n}_Mod{m}', incr, has_ce, has_reset, cin, cout)
    args = []
    if cin:
        args += ['CIN', In(Bit)]

    args += ["O", Out(UInt[n])]
    if cout:
        args += ["COUT", Out(Bit)]

    args += ClockInterface(has_ce, has_reset)

    class CounterModM(Circuit):
        name = name_
        io = IO(**dict(zip(args[::2], args[1::2])))

        counter = Counter(n, cin=cin, cout=False, incr=incr,
                       has_ce=has_ce, has_reset=True)
        reset = Decode(m - 1, n)(counter.O)

        if has_reset:
            reset = Or(2)(reset, io.RESET)

        if has_ce:
            CE = io.CE
            reset = And(2)(reset, CE)
            # reset is sometimes called rollover or RO
            # note that we don't return RO in Counter

            # should also handle r in the definition

        wire(reset, counter.RESET) # synchronous reset

        if has_ce:
            wire(CE, counter.CE)

        if cin:
            wire( io.CIN, counter.CIN )

        wire( counter.O, io.O )

        if cout:
            wire( reset, io.COUT )

    return CounterModM

def CounterModM(m, n, cin=False, cout=True, incr=1,
    has_ce=False, has_reset=False, **kwargs):
    return DefineCounterModM(m, n, cin, cout, incr, has_ce, has_reset=has_reset)(**kwargs)

def SizedCounterModM(m, cin=False, cout=False, incr=1,
    has_ce=False, has_reset=False, **kwargs):
    """
    This is that counts from 0 to m - 1 that uses the minimum number of bits
    :param m: The value the counter counts up to
    :param cin: Whether this counter should have a carry input
    :param cout: Whether this counter should a carry output
    :param incr: How much this counter should increment by per clock. Default is 1.
    :param has_ce: Whether this counter should a clock-enable input
    :param kwargs: Args passed to the counter circuit when it is being initialized
    :return: A counter circuit
    """
    return DefineCounterModM(m, math.ceil(math.log(m, 2)), cin, cout, incr, has_ce, has_reset)(**kwargs)

DefineUpCounterModM = DefineCounterModM
UpCounterModM = CounterModM



from magma import Circuit, In, Out, Bits, wire, bits, uncurry, fork, cache_definition
from magma.bitutils import int2seq
from mantle import LUT
from mantle import Mux
from .RAM import readport

__all__  = ['DefineROM', 'ROM']

def ROM4(data, i, width):
    return fork([uncurry(LUT(data[i][w], 4)) for w in range(width)])

def ROM4s(n, width, data):
    return [ROM4(data, i, width) for i in range(n//16)]

def MUXs(n, width):
    return [Mux(2,width) for i in range(n)]

def interleave16(data, width):
    n = len(data)
    bits = [int2seq(data[i], width) for i in range(n)]
    #print(n, bits)
    data = [ [ [bits[i+j][w] for j in range(16)] for w in range(width)] \
               for i in range(0,n,16) ]
    #print(len(data), data)
    return data

def DefineROM(height, width, data):
    assert height >= 4
    n = 1 << height

    data = interleave16(data, width)

    TADDR = Bits(height)
    TDATA = Bits(width)

    class _ROM(Circuit):
        name = f'ROM{n}x{width}'
        IO = ['RADDR', In(TADDR), 'RDATA', Out(TDATA)]

        @classmethod
        def definition(io):
            roms = ROM4s(n, width, data)
            [roms[i](io.RADDR[0:4]) for i in range(n//16)]
            wire( readport(height-4, width, roms, io.RADDR[4:]), io.RDATA )

    return _ROM

def ROM(height, width, data=None):
    return DefineROM(height, width, data=data)()




import magma as m
from hwtypes import SIntVector, UIntVector


_default_port_mapping = {
    "I": "in",
    "I0": "in0",
    "I1": "in1",
    "O": "out",
    "S": "sel",
    "CLK": "clk",
}


def DeclareCoreirCircuit(*args, **kwargs):
    name_ = args[0]
    decl = args[1:]
    assert isinstance(name_, str)
    assert len(decl) % 2 == 0

    class _Circuit(m.Circuit):
        name = name_
        io = m.IO(**dict(zip(decl[::2], decl[1::2])))
        renamed_ports = _default_port_mapping

    kwargs.setdefault("primitive", True)
    kwargs.setdefault("stateful", False)
    kwargs.setdefault("_is_definition", False)
    for k, v in kwargs.items():
        setattr(_Circuit, k, v)

    return _Circuit


def get_int_vector_type(signed):
    return SIntVector if signed else UIntVector



from magma import *
from .simulation import *

__all__  = ['A0', 'A1', 'A2', 'A3']
__all__ += ['I0', 'I1', 'I2', 'I3']
__all__ += ['ALL', 'ANY', 'PARITY']
__all__ += ['ZERO', 'ONE']
__all__ += ['LUTS_PER_LOGICBLOCK', 'BITS_PER_LUT', 'LOG_BITS_PER_LUT']

__all__ += ['SB_LUT4']
__all__ += ['SB_CARRY']

__all__ += ['SB_DFF',    'SB_DFFE',
            'SB_DFFSR',  'SB_DFFESR',
            'SB_DFFSS',  'SB_DFFESS',
            'SB_DFFR',   'SB_DFFER',
            'SB_DFFS',   'SB_DFFES',
            'SB_DFFN',   'SB_DFFNE',
            'SB_DFFNSR', 'SB_DFFNESR',
            'SB_DFFNSS', 'SB_DFFNESS',
            'SB_DFFNR',  'SB_DFFNER',
            'SB_DFFNS',  'SB_DFFNES']

LUTS_PER_LOGICBLOCK = 2
LOG_BITS_PER_LUT = 4
BITS_PER_LUT = 1 << LOG_BITS_PER_LUT

ZERO = 0
ONE = (1<<BITS_PER_LUT)-1

A0 = 0xAAAA
A1 = 0xCCCC
A2 = 0xF0F0
A3 = 0xFF00

I0 = A0
I1 = A1
I2 = A2
I3 = A3

ALL = A0 & A1 & A2 & A3
ANY = A0 | A1 | A2 | A3
PARITY = A0 ^ A1 ^ A2 ^ A3

SB_LUT4 = DeclareCircuit('SB_LUT4',
               "I0", In(Bit),
               "I1", In(Bit),
               "I2", In(Bit),
               "I3", In(Bit),
               "O",  Out(Bit),
               stateful=False,
               simulate=simulate_sb_lut4,
               coreir_lib="ice40")

# Implements (I0&I1)|(I1&I2)|(I2&I0)
SB_CARRY = DeclareCircuit('SB_CARRY',
               "I0", In(Bit), # must be the same as SB_LUT4 I1 to pack
               "I1", In(Bit), # must be the same as SB_LUT4 I2 to pack
               "CI", In(Bit), # must be from previous SB_LUT4 to pack
               "CO", Out(Bit),
               stateful=False,
               simulate=simulate_sb_carry,
               coreir_lib="ice40")

# Positive edge versions

# DFF - D Flip-Flop
SB_DFF = DeclareCircuit('SB_DFF',
               "C", In(Clock),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=False, r=False, s=False, n=False),
               coreir_lib="ice40")

# DFF w/ Clock enable
SB_DFFE = DeclareCircuit('SB_DFFE',
               "C", In(Clock),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=False, r=False, s=False, n=False),
               coreir_lib="ice40")

# DFF w/ Synchronous Reset
SB_DFFSR = DeclareCircuit('SB_DFFSR',
               "C", In(Clock),
               "R", In(Reset),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=True, r=True, s=False, n=False),
               coreir_lib="ice40")

# DFF w/ Asynchronous Reset
SB_DFFR = DeclareCircuit('SB_DFFR',
               "C", In(Clock),
               "R", In(Reset),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=False, r=True, s=False, n=False),
               coreir_lib="ice40")

# DFF w/ Synchronous Set
SB_DFFSS = DeclareCircuit('SB_DFFSS',
               "C", In(Clock),
               "S", In(Bit),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=True, r=False, s=True, n=False),
               coreir_lib="ice40")

# DFF w/ Asynchronous Set
SB_DFFS = DeclareCircuit('SB_DFFS',
               "C", In(Clock),
               "S", In(Bit),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=False, r=False, s=True, n=False),
               coreir_lib="ice40")

# DFF w/ Synchronous Reset, Clock enable
SB_DFFESR = DeclareCircuit('SB_DFFESR',
               "C", In(Clock),
               "R", In(Reset),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=True, r=True, s=False, n=False),
               coreir_lib="ice40")

# DFF w/ Asynchronous Reset, Clock enable
SB_DFFER = DeclareCircuit('SB_DFFER',
               "C", In(Clock),
               "R", In(Reset),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=False, r=True, s=False, n=False),
               coreir_lib="ice40")

# DFF w/ Synchronous Set, Clock enable
SB_DFFESS = DeclareCircuit('SB_DFFESS',
               "C", In(Clock),
               "S", In(Bit),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=True, r=False, s=True, n=False),
               coreir_lib="ice40")

# DFF w/ Asynchronous Set, Clock enable
SB_DFFES = DeclareCircuit('SB_DFFES',
               "C", In(Clock),
               "S", In(Bit),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=False, r=False, s=True, n=False),
               coreir_lib="ice40")

# Negative edge versions
SB_DFFN = DeclareCircuit('SB_DFFN',
               "C", In(Clock),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=False, r=False, s=False, n=True),
               coreir_lib="ice40")

# DFFN w/ Clock enable
SB_DFFNE = DeclareCircuit('SB_DFFNE',
               "C", In(Clock),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=False, r=False, s=False, n=True),
               coreir_lib="ice40")

# DFFN w/ Synchronous Reset
SB_DFFNSR = DeclareCircuit('SB_DFFNSR',
               "C", In(Clock),
               "R", In(Reset),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=True, r=True, s=False, n=True),
               coreir_lib="ice40")

# DFFN w/ Asynchronous Reset
SB_DFFNR = DeclareCircuit('SB_DFFNR',
               "C", In(Clock),
               "R", In(Reset),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=False, r=True, s=False, n=True),
               coreir_lib="ice40")

# DFFN w/ Synchronous Set
SB_DFFNSS = DeclareCircuit('SB_DFFNSS',
               "C", In(Clock),
               "S", In(Bit),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=True, r=False, s=True, n=True),
               coreir_lib="ice40")

# DFFN w/ Asynchronous Set
SB_DFFNS = DeclareCircuit('SB_DFFNS',
               "C", In(Clock),
               "S", In(Bit),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=False, sy=False, r=False, s=True, n=True),
               coreir_lib="ice40")

# DFFN w/ Synchronous Reset, Clock enable
SB_DFFNESR = DeclareCircuit('SB_DFFNESR',
               "C", In(Clock),
               "R", In(Reset),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simualte=gen_sb_dff_sim(ce=True, sy=True, r=True, s=False, n=True),
               coreir_lib="ice40")

# DFFN w/ Asynchronous Reset, Clock enable
SB_DFFNER = DeclareCircuit('SB_DFFNER',
               "C", In(Clock),
               "R", In(Reset),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=False, r=True, s=False, n=True),
               coreir_lib="ice40")

# DFFN w/ Synchronous Set, Clock enable
SB_DFFNESS = DeclareCircuit('SB_DFFNESS',
               "C", In(Clock),
               "S", In(Bit),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=True, r=False, s=True, n=True),
               coreir_lib="ice40")

# DFFN w/ Asynchronous Set, Clock enable
SB_DFFNES = DeclareCircuit('SB_DFFNES',
               "C", In(Clock),
               "S", In(Bit),
               "E", In(Enable),
               "D", In(Bit),
               "Q", Out(Bit),
               stateful=True,
               simulate=gen_sb_dff_sim(ce=True, sy=False, r=False, s=True, n=True),
               coreir_lib="ice40")




from magma import *
import magma as m
from .LUT import LUT3, A0, A1, A2

__all__  = ['Mux2', 'Mux4', 'Mux8', 'Mux16']
__all__ += ['DefineMux', 'Mux']

#
# C ? B : A
#
MUX2DATA = (~A2&A0)|(A2&A1)

"""Construct a Mux with 2 1-bit inputs."""
class Mux2(Circuit):
    IO = ["I", In(Bits[ 2 ]), "S", In(Bit), "O", Out(Bit) ]
    @classmethod
    def definition(mux2):
        lut = LUT3(MUX2DATA)
        wire(mux2.I[0], lut.I0)
        wire(mux2.I[1], lut.I1)
        wire(mux2.S, lut.I2)
        wire(lut.O, mux2.O)


# """Construct a Mux with 4 1-bit inputs."""
class Mux4(Circuit):
    IO = ["I", In(Bits[ 4 ]), "S", In(Bits[ 2 ]), "O", Out(Bit) ]
    @classmethod
    def definition(mux4):
        mux0 = Mux2()
        mux1 = Mux2()
        mux = Mux2()
        mux0(mux4.I[0:2],mux4.S[0])
        mux1(mux4.I[2:4],mux4.S[0])
        mux( array( [mux0.O, mux1.O] ), mux4.S[1] )
        wire( mux.O, mux4.O )

# """Construct a Mux with 8 1-bit inputs."""
class Mux8(Circuit):
    IO = ["I", In(Bits[ 8 ]), "S", In(Bits[ 3 ]), "O", Out(Bit) ]
    @classmethod
    def definition(mux8):
        mux0 = Mux4()
        mux1 = Mux4()
        mux = Mux2()
        mux0(mux8.I[0:4], mux8.S[0:2])
        mux1(mux8.I[4:8], mux8.S[0:2])
        mux( array( [mux0.O, mux1.O] ), mux8.S[2] )
        wire( mux.O, mux8.O )

# """Construct a Mux with 16 1-bit inputs."""
class Mux16(Circuit):
    IO = ["I", In(Bits[ 16 ]), "S", In(Bits[ 4 ]), "O", Out(Bit) ]
    @classmethod
    def definition(mux16):
        mux0 = Mux8()
        mux1 = Mux8()
        mux = Mux2()
        mux0(mux16.I[0:8],  mux16.S[0:3])
        mux1(mux16.I[8:16], mux16.S[0:3])
        mux( array( [mux0.O, mux1.O] ), mux16.S[3])
        wire(mux.O, mux16.O)

def MuxN(height, **kwargs):
    assert height in [2, 4, 8, 16]

    if height == 2:
        return Mux2(**kwargs)
    elif height == 4:
        return Mux4(**kwargs)
    elif height == 8:
        return Mux8(**kwargs)
    elif height == 16:
        return Mux16(**kwargs)

# Mux Module name
def _MuxName(height, width):
    return f'Mux{height}x{width}'

def _MuxInterface(height, width):
    AW = In(Bits[ width ])
    if   height == 2:
        args = ['I0', AW,
                'I1', AW]
        args += ['S', In(Bit)]
    elif height == 4:
        args = ['I0', AW,
                'I1', AW,
                'I2', AW,
                'I3', AW]
        args += ['S', In(Bits[ 2 ])]
    elif height == 8:
        args = ['I0', AW,
                'I1', AW,
                'I2', AW,
                'I3', AW,
                'I4', AW,
                'I5', AW,
                'I6', AW,
                'I7', AW]
        args += ['S', In(Bits[ 3 ])]
    elif height == 16:
        args = ['I0',  AW,
                'I1',  AW,
                'I2',  AW,
                'I3',  AW,
                'I4',  AW,
                'I5',  AW,
                'I6',  AW,
                'I7',  AW,
                'I8',  AW,
                'I9',  AW,
                'I10', AW,
                'I11', AW,
                'I12', AW,
                'I13', AW,
                'I14', AW,
                'I15', AW]
        args += ['S', In(Bits[ 4 ])]

    args += ['O', Out(AW)]

    return args


def DefineMux(height=2, width=1):

    """
    Construct a Mux. Height inputs are width bits wide.
    """

    assert height in [2, 4, 8, 16]
    if width is None:
        if height == 2:
            return Mux2
        elif height == 4:
            return Mux4
        elif height == 8:
            return Mux8
        elif height == 16:
            return Mux16

    class _Mux(Circuit):
        name = _MuxName(height, width)
        IO = _MuxInterface(height, width)
        @classmethod
        def definition(Mux):
            def amux(y):
                if height == 2:
                    return curry(MuxN(height, loc=(0,y//8,y%8)), prefix='I')
                return curry(MuxN(height), prefix='I')
            mux = braid(col(amux, width), forkargs=['S'])

            if   height == 2:  mux( Mux.I0, Mux.I1, Mux.S )
            elif height == 4:  mux( Mux.I0, Mux.I1, Mux.I2, Mux.I3, Mux.S )
            elif height == 8:  mux( Mux.I0, Mux.I1, Mux.I2, Mux.I3,
                                    Mux.I4, Mux.I5, Mux.I6, Mux.I7, Mux.S )
            elif height == 16: mux( Mux.I0, Mux.I1, Mux.I2, Mux.I3,
                                    Mux.I4, Mux.I5, Mux.I6, Mux.I7,
                                    Mux.I8, Mux.I9, Mux.I10, Mux.I11,
                                    Mux.I12, Mux.I13, Mux.I14, Mux.I15, Mux.S )
            wire( mux.O, Mux.O )
    return _Mux

def Mux(height=2, width=None, T=None, **kwargs):
    if T is not None:
        assert width is None, "Can only specify width **or** T"
        if issubclass(T, m.Bit):
            width = None
        elif issubclass(T, m.Bits) or \
                issubclass(T, m.Array) and issubclass(T.T, m.Bit):
            width = len(T)
        else:
            raise NotImplementedError(type(T))
    if width is None:
        return curry(MuxN(height, **kwargs), prefix="I")

    return DefineMux(height, width)(**kwargs)




import magma as m
from .FF import DefineCoreirReg, DefineDFF
import mantle


@m.cache_definition
def DefineRegister(n, init=0, has_ce=False, has_reset=False,
                   has_async_reset=False, has_async_resetn=False, _type=m.Bits,
                   reset_priority=True, ):
    if has_reset and (has_async_reset or has_async_resetn):
        raise ValueError("Cannot have synchronous and asynchronous reset")
    if has_async_resetn and has_async_reset:
        raise ValueError("Cannot have posedge and negedge asynchronous reset")

    if has_reset or has_ce:
        if n is None:
            T = m.Bit
        else:
            T = _type[n]
        class Register(m.Circuit):
            name = f"Register_has_ce_{has_ce}_has_reset_{has_reset}_" \
                   f"has_async_reset_{has_async_reset}_" \
                   f"has_async_resetn_{has_async_resetn}_" \
                   f"type_{_type.__name__}_n_{n}"
            io = m.IO(I=m.In(T), O=m.Out(T))
            io += m.ClockIO(has_ce=has_ce, has_reset=has_reset,
                            has_async_reset=has_async_reset,
                            has_async_resetn=has_async_resetn)

            reg = DefineCoreirReg(n, init, has_async_reset,
                                  has_async_resetn, _type)(name="value")
            I = io.I
            O = reg.O
            if n is None:
                O = O[0]
            if has_reset and has_ce:
                if reset_priority:
                    I = mantle.mux([O, I], io.CE, name="enable_mux")
                    I = mantle.mux([I, m.bits(init, n)], io.RESET)
                else:
                    I = mantle.mux([I, m.bits(init, n)], io.RESET)
                    I = mantle.mux([O, I], io.CE, name="enable_mux")
            elif has_ce:
                I = mantle.mux([O, I], io.CE, name="enable_mux")
            elif has_reset:
                I = mantle.mux([I, m.bits(init, n)], io.RESET)
            if n is None:
                m.wire(I, reg.I[0])
            else:
                m.wire(I, reg.I)
            m.wire(io.O, O)

        return Register
    elif n is None:
        if _type is not m.Bits:
            raise NotImplementedError()
        return DefineDFF(init, has_ce, has_async_reset=has_async_reset)
    else:
        return DefineCoreirReg(n, init, has_async_reset, has_async_resetn,
                               _type)


def Register(n, init=0, has_ce=False, has_reset=False, has_async_reset=False,
             has_async_resetn=False, _type = m.Bits, **kwargs):
    return DefineRegister(n, init, has_ce, has_reset, has_async_reset,
                          has_async_resetn=has_async_resetn, _type=_type)(**kwargs)


def register(I, ce=None, reset=None, async_reset=None, **kwargs):
    has_ce = ce is not None
    has_reset = reset is not None
    has_async_reset = async_reset is not None
    reg = Register(len(I),
                   has_ce=has_ce,
                   has_reset=has_reset,
                   has_async_reset=has_async_reset,
                   **kwargs)
    reg(I)
    if has_ce:
        m.wire(ce, reg.CE)
    if has_reset:
        m.wire(reset, reg.RESET)
    if has_async_reset:
        m.wire(reset, reg.ASYNCRESET)
    return reg.O



from magma import *
from ..ice40.PLB import SB_CARRY, A0, A1, A2, A3
from .LUT import LUT2

__all__  = ["HalfAdder", 'halfadder']

class HalfAdder(Circuit):
    IO = ["I0", In(Bit), "I1", In(Bit), "O", Out(Bit), "COUT", Out(Bit)]
    @classmethod
    def definition(io):
        sum = LUT2(A0^A1)
        carry = SB_CARRY() # (A0&A1)|(A1&A2)|(A2&A0)
        wire( sum(io.I0,io.I1), io.O )
        wire( carry(io.I0,io.I1,0), io.COUT )

def halfadder(a, b, **kwargs):
    return HalfAdder()(a, b, **kwargs)




from magma import *
import magma as m
from mantle.coreir.MUX import Mux
import coreir
from hwtypes import BitVector

from .util import DeclareCoreirCircuit

def gen_sim_register(N, init, has_ce, has_async_reset, has_async_resetn):
    def sim_register(self, value_store, state_store):
        """
        Adapted from Brennan's SB_DFF simulation in mantle
        """
        if has_async_resetn and has_async_reset:
            raise ValueError("Cannot have posedge and negedge asynchronous reset")
        cur_clock = value_store.get_value(self.CLK)

        if not state_store:
            state_store['prev_clock'] = cur_clock
            state_store['cur_val'] = BitVector[N](init) if N is not None else bool(init)

        if has_async_reset or has_async_resetn:
            cur_reset = value_store.get_value(self.arst)
        # if s:
        #     cur_s = value_store.get_value(self.S)

        prev_clock = state_store['prev_clock']
        # if not n:
        #     clock_edge = cur_clock and not prev_clock
        # else:
        #     clock_edge = not cur_clock and prev_clock
        clock_edge = cur_clock and not prev_clock

        new_val = state_store['cur_val'].as_bool_list() if N is not None else state_store['cur_val']

        if clock_edge:
            new_val = value_store.get_value(self.I)

        if has_async_reset and cur_reset or has_async_resetn and not cur_reset:
            new_val = BitVector[N](init) if N is not None else bool(init)
        # if s and not sy and cur_s:
        #     new_val = True

        state_store['prev_clock'] = cur_clock
        state_store['cur_val'] = BitVector[N](new_val) if N is not None else new_val
        value_store.set_value(self.O, new_val)
    return sim_register


@m.cache_definition
def DefineCoreirReg(width, init=0, has_async_reset=False,
                    has_async_resetn=False, T=Bits):
    if width is None:
        width = 1
    name = "reg_P"  # TODO: Add support for clock interface
    config_args = {"init": coreir.type.BitVector[width](init)}
    gen_args = {"width": width}
    T = T[width]
    io = ["I", In(T), "CLK", In(Clock), "O", Out(T)]
    methods = []

    def reset(self, condition):
        wire(condition, self.rst)
        return self

    if has_async_resetn and has_async_reset:
        raise ValueError("Cannot have posedge and negedge asynchronous reset")
    if has_async_reset:
        io.extend(["arst", In(AsyncResetIn)])
        name += "R"  # TODO: This assumes ordering of clock parameters
        config_args["arst_posedge"] = True
    if has_async_resetn:
        io.extend(["arst", In(AsyncResetNIn)])
        name += "R"  # TODO: This assumes ordering of clock parameters
        config_args["arst_posedge"] = False

    def when(self, condition):
        wire(condition, self.en)
        return self

    # if has_ce:
    #     io.extend(["en", In(Enable)])
    #     name += "E"  # TODO: This assumes ordering of clock parameters
    #     methods.append(circuit_type_method("when", when))
    #     gen_args["has_en"] = True

    # default_kwargs = gen_args.copy()
    default_kwargs = {"init": coreir.type.BitVector[width](init)}
    # default_kwargs.update(config_args)

    coreir_name = "reg_arst" if (has_async_reset or has_async_resetn) else "reg"

    return DeclareCoreirCircuit(
        name,
        *io,
        stateful=True,
        simulate=gen_sim_register(width, init, False, has_async_reset,
                                  has_async_resetn),
        circuit_type_methods=methods,
        default_kwargs=default_kwargs,
        coreir_genargs=gen_args,
        coreir_configargs=config_args,
        coreir_name=coreir_name,
        verilog_name="coreir_" + coreir_name,
        coreir_lib="coreir"
    )

def define_wrap(type_, type_name, in_type):

    def _sim_wrap(self, value_store, state_store):
        input_val = value_store.get_value(getattr(self, "in"))
        value_store.set_value(self.out, input_val)

    args = {"in": In(in_type), "out": Out(type_)}

    class _coreir_wrap(Circuit):
        name = f'coreir_wrap{type_name}'
        io = IO(**args)
        coreir_genargs = {"type": AsyncReset}
        coreir_name = "wrap"
        coreir_lib = "coreir"
        simulate = _sim_wrap

    return _coreir_wrap

@m.cache_definition
def DefineDFF(init=0, has_ce=False, has_reset=False, has_async_reset=False, has_async_resetn=False):
    Reg = DefineCoreirReg(None, init, has_async_reset, has_async_resetn)
    class _DFF(m.Circuit):
        io = m.IO(I=In(Bit), O=Out(Bit))
        io += m.ClockIO(has_ce=has_ce, has_reset=has_reset,
                        has_async_reset=has_async_reset,
                        has_async_resetn=has_async_resetn)
        name = "DFF_init{}_has_ce{}_has_reset{}_has_async_reset{}".format(
            init, has_ce, has_reset, has_async_reset, has_async_resetn)

        value = Reg()
        I = io.I
        if has_reset and (has_async_reset or has_async_resetn):
            raise ValueError("Cannot have synchronous and asynchronous reset")
        if has_async_resetn and has_async_reset:
            raise ValueError("Cannot have posedge and negedge asynchronous reset")
        if has_reset:
            I = Mux()(io.I, bit(init), io.RESET)
        if has_ce:
            I = Mux()(value.O[0], I, io.CE)
        wire(I, value.I[0])
        wire(value.O[0], io.O)
    return _DFF




def DFF(init=0, has_ce=False, has_reset=False, has_async_reset=False, **kwargs):
    return DefineDFF(init, has_ce, has_reset, has_async_reset)(**kwargs)

FF = DFF



from magma import Circuit, Bits, In, Out, wire, fork, join, flat, uncurry, map_, cache_definition
from mantle import And, Or

__all__  = ['Swap',  'swap']
__all__ += ['Swaps', 'swaps']
__all__ += ['DefineEvenOddSwaps', 'EvenOddSwaps', 'evenoddswaps']

#
# Binary swap circuit
#
class Swap(Circuit):
    IO = ['I', In(Bits(2)), "O", Out(Bits(2))]
    @classmethod
    def definition(io):
        swap = uncurry( fork( And(2), Or(2) ) , prefix="I")
        #swap = uncurry( fork( And(2), Or(2) ) , prefix="in")
        wire( swap( io.I ), io.O )

def swap(I):
    return Swap()(I)


def DefineSwaps(n):
    class Swaps(Circuit):
        name = 'Swap{}'.format(n)
        IO = ['I', In(Bits(n)), "O", Out(Bits(n))]
        @classmethod
        def definition(io):
            s = flat( join( map_(Swap, n//2) ), flatargs = ['I', 'O'] )
            wire(s(io.I), io.O)
    return Swaps

def Swaps(n):
    return DefineSwaps(n)()

def swaps(I):
    return Swaps(len(I))(I)


def DefineEvenOddSwaps(n):
    class EvenOddSwaps(Circuit):
        name = 'EvenOddSwap{}'.format(n)
        IO = ['I', In(Bits(n)), "O", Out(Bits(n))]
        @classmethod
        def definition(io):
            s = flat( join( map_(Swap, n//2-1) ), flatargs = ['I', 'O'] )
            wire(io.I[0], io.O[0])
            wire(s(io.I[1:-1]), io.O[1:-1])
            wire(io.I[-1], io.O[-1])
    return EvenOddSwaps

def EvenOddSwaps(n):
    return DefineEvenOddSwaps(n)()

def evenoddswaps(I):
    return EvenOddSwaps(len(I))(I)



from collections import OrderedDict
from magma import *
from .simulation import gen_sb_ram40_4k_sim

__all__  = ['SB_RAM40_4K']

__all__ += ['RAMB', 'ROMB']

# posedge read clock, posedge write clock
SB_RAM40_4K = DeclareCircuit("SB_RAM40_4K",
    "RDATA", Out(Bits[ 16 ]),
    "RADDR", In(Bits[ 11 ]),
    "RCLK",  In(Clock),
    "RCLKE", In(Enable),
    "RE",    In(Enable),
    "WCLK",  In(Clock),
    "WCLKE", In(Enable),
    "WE",    In(Enable),
    "WADDR", In(Bits[ 11 ]),
    "MASK",  In(Bits[ 16 ]),
    "WDATA", In(Bits[ 16 ]),
    stateful=True,
    simulate=gen_sb_ram40_4k_sim(prc=True, pwc=True),
    coreir_lib='ice40')

# posedge read clock, negedge write clock
# negedge read clock, posedge write clock
# negedge read clock, negedge write clock
# - can get these effects by inverting the RCLK and/or the WCLK
# NEGCLK_R(0|1)
# NEGCLK_W(0|1)


# MASK=0, write-enabled, MASK=1, write-disabled

#
# N is the number of bits of output
#
def init(rom,N,mode=0):
    # INIT_%x 256 bits = 32 bytes = 64 nibbles
    params = OrderedDict({})
    params['WRITE_MODE'] = mode
    params['READ_MODE'] = mode

    # M is the number of high (>8) address values
    #  e.g. if N == 8, then there are 2 high address values
    M = 16//N
    for i in range(16):
        v = 0
        for b in range(256):
            col = (b//M)%N
            # NB. the high address values ocuppy the lowest bit positions
            row = 256*(b%M) + 16*i + b//16
            bit = (rom[row] >> col)&1
            #print('i=%x'%i, 'b=%d'%b, row, col)
            v |= bit << b
        key = 'INIT_%X' % i
        #params[key] = "256'h%064x" % v
        params[key] = (v, 256)
    return params

def wireaddr(addr, n):
    for i in range(n):
        wire(0, addr[10-i])
    return addr[0:11-n]

def _RAMB(height, width, rom=None, readonly=False):
    if not rom:
        rom = height * [0]
    n = len(rom)
    assert height == n

    ram40 = None
    if   n ==  256: # 256x16
        assert width == 16
        params = init(rom,16,mode=0)
        ram40 = SB_RAM40_4K(**params)
        ram40.RADDR = wireaddr(ram40.RADDR, 3)
        if not readonly:
            ram40.WADDR = wireaddr(ram40.WADDR, 3)
    elif n ==  512: # 512x8
        assert width == 8
        params = init(rom,8,mode=1)
        ram40 = SB_RAM40_4K(**params)
        ram40.RADDR = wireaddr(ram40.RADDR, 2)
        RDATA = ram40.RDATA
        ram40.RDATA = array([RDATA[0], RDATA[2], RDATA[4], RDATA[6],
                            RDATA[8], RDATA[10], RDATA[12], RDATA[14]])
        if not readonly:
            WDATA = ram40.WDATA
            wire(0,WDATA[1])
            wire(0,WDATA[3])
            wire(0,WDATA[5])
            wire(0,WDATA[7])
            wire(0,WDATA[9])
            wire(0,WDATA[11])
            wire(0,WDATA[13])
            wire(0,WDATA[15])
            ram40.WDATA = array([WDATA[0], WDATA[2], WDATA[4], WDATA[6],
                                WDATA[8], WDATA[10], WDATA[12], WDATA[14]])
            ram40.WADDR = wireaddr(ram40.WADDR, 2)
    elif n == 1024: # 1024x4
        assert width == 4
        params = init(rom,4,mode=2)
        ram40 =  SB_RAM40_4K(**params)
        ram40.RADDR = wireaddr(ram40.RADDR, 1)
        RDATA = ram40.RDATA
        ram40.RDATA = array([RDATA[1], RDATA[5], RDATA[9], RDATA[13]])
        if not readonly:
            WDATA = ram40.WDATA
            wire(0,WDATA[0])
            wire(0,WDATA[2])
            wire(0,WDATA[3])
            wire(0,WDATA[4])
            wire(0,WDATA[6])
            wire(0,WDATA[7])
            wire(0,WDATA[8])
            wire(0,WDATA[10])
            wire(0,WDATA[11])
            wire(0,WDATA[12])
            wire(0,WDATA[14])
            wire(0,WDATA[15])
            ram40.WDATA = array([WDATA[1], WDATA[5], WDATA[9], WDATA[13]])
            ram40.WADDR = wireaddr(ram40.WADDR, 1)
    elif n == 2048: # 2048x2
        assert width == 2
        params = init(rom,2,mode=3)
        ram40 = SB_RAM40_4K(**params)
        RDATA = ram40.RDATA
        ram40.RDATA = array([RDATA[3], RDATA[11]])
        if not readonly:
            WDATA = ram40.WDATA
            wire(0,WDATA[0])
            wire(0,WDATA[1])
            wire(0,WDATA[2])
            wire(0,WDATA[4])
            wire(0,WDATA[5])
            wire(0,WDATA[6])
            wire(0,WDATA[7])
            wire(0,WDATA[8])
            wire(0,WDATA[9])
            wire(0,WDATA[10])
            wire(0,WDATA[12])
            wire(0,WDATA[13])
            wire(0,WDATA[14])
            wire(0,WDATA[15])
            ram40.WDATA = array([WDATA[3], WDATA[11]])

    if readonly:
        wire( 0, ram40.WE    )
        wire( array(11*[0]), ram40.WADDR )
        wire( array(16*[0]), ram40.WDATA )
    wire(0 if readonly else 1, ram40.WCLKE )
    wire(1, ram40.RCLKE)
    wire(array(16*[0]), ram40.MASK)
    if readonly:
        return AnonymousCircuit("RADDR", ram40.RADDR,
                                "RDATA", ram40.RDATA,
                                "RCLK",  ram40.RCLK,
                                "RE",    ram40.RE)
    else:
        return AnonymousCircuit("RADDR", ram40.RADDR,
                                "RDATA", ram40.RDATA,
                                "RCLK",  ram40.RCLK,
                                "RE",    ram40.RE,
                                "WADDR", ram40.WADDR,
                                "WDATA", ram40.WDATA,
                                "WCLK",  ram40.WCLK,
                                "WE",    ram40.WE)

def RAMB(height, width, ram=None):
    return _RAMB(height, width, ram, readonly=False)

def ROMB(height, width, rom=None):
    return _RAMB(height, width, rom, readonly=True)



from magma import *
from mantle import And, DefineAdd
from mantle import Mux
from .register import Register
from .counter import counter_name

__all__ = ['DefineCounterLoad', 'CounterLoad']

#
# Create an n-bit counter with increment and load
#
#   DATA : In(UInt(n)), LOAD : In(Bit), O : Out(UInt(n)), COUT : Out(Bit)
#
def DefineCounterLoad(n, cin=False, cout=True, incr=1, has_ce=False, has_reset=False):

    name_ = counter_name(f'CounterLoad{n}', incr, has_ce, has_reset, cin, cout)

    args = []
    args += ['DATA', In(UInt[ n ])]
    args += ['LOAD', In(Bit)]
    if cin:
        args += ['CIN', In(Bit)]

    args += ["O", Out(UInt[ n ])]
    if cout:
        args += ["COUT", Out(Bit)]

    args += ClockInterface(has_ce, has_reset)

    class Counter(Circuit):
        name = name_
        io = IO(**dict(zip(args[::2], args[1::2])))

        add = DefineAdd(n, cin=cin, cout=cout)()
        mux = Mux(2, n)
        reg = Register(n, has_ce=has_ce, has_reset=has_reset)

        wire( reg.O, add.I0 )
        wire( array(incr, n), add.I1 )

        wire( add.O, mux.I0 )
        wire( io.DATA, mux.I1 )
        wire( io.LOAD, mux.S )

        reg(mux.O)

        next = False
        if next:
            wire( mux.O, io.O )
        else:
            wire( reg.O, io.O )

        if cin:
            wire( io.CIN, add.CIN )

        if cout:
            wire( add.COUT, io.COUT ) # this is fishy because of the LOAD

    wireclock(Counter, Counter.reg)

    return Counter

def CounterLoad(n, cin=False, cout=True, incr=1,
        has_ce=False, has_reset=False, **kwargs):
    """Construct a n-bit counter."""
    return DefineCounterLoad(n, cin=cin, cout=cout, incr=incr,
               has_ce=has_ce, has_reset=has_reset)(**kwargs)



import magma as m
from mantle.coreir import DefineRegister


def _make_read_type(data_width, addr_width):
    fields = dict(data=m.Out(m.Bits[data_width]), addr=m.In(m.Bits[addr_width]))
    return m.Product.from_fields("ReadPortT", fields)


def _make_write_type(data_width, addr_width):
    fields = dict(data=m.In(m.Bits[data_width]), addr=m.In(m.Bits[addr_width]))
    return m.Product.from_fields("WritePortT", fields)


class RegFileBuilder(m.CircuitBuilder):
    def __init__(self, name, height: int, width: int, backend: str = "magma"):
        super().__init__(name)
        self._data_width = width
        self._height = height
        self._addr_width = m.bitutils.clog2(height)
        self._read_ports = []
        self._write_ports = []
        self._enable_ports = {}
        self._readT = _make_read_type(self._data_width, self._addr_width)
        self._writeT = _make_write_type(self._data_width, self._addr_width)
        clocks = m.ClockIO(has_async_reset=True).decl()
        for name, typ in zip(clocks[::2], clocks[1::2]):
            self._add_port(name, typ)
        self.backend = backend

    @m.builder_method
    def __getitem__(self, addr):
        """Add a read port"""
        count = len(self._read_ports)
        name = f"read_{count}"
        self._add_port(name, self._readT)
        self._read_ports.append(name)
        port = getattr(self, name)
        m.wire(addr, port.addr)
        return port.data

    def _add_write_port(self, addr, data):
        """Add a write port"""
        count = len(self._write_ports)
        name = f"write_{count}"
        self._add_port(name, self._writeT)
        self._write_ports.append(name)
        port = getattr(self, name)
        m.wire(addr, port.addr)
        m.wire(data, port.data)
        return name

    @m.builder_method
    def __setitem__(self, addr, data):
        self._add_write_port(addr, data)

    def _add_enable(self, port_name, enable):
        enable_name = f"{port_name}_en"
        self._add_port(enable_name, m.In(m.Enable))
        self._enable_ports[port_name] = enable_name
        port = getattr(self, enable_name)
        m.wire(enable, port)

    @m.builder_method
    def write(self, addr, data, enable=None):
        name = self._add_write_port(addr, data)
        if enable is not None:
            self._add_enable(name, enable)

    def _finalize_magma(self):
        registers = [DefineRegister(self._data_width, has_async_reset=True)()
                     for _ in range(self._height)]
        read_data = {name: None for name in self._read_ports}
        reg_data = [reg.O for reg in registers]
        for name in self._read_ports:
            port = self._port(name)
            mux = m.Mux(self._height, m.Bits[self._data_width])()
            read_data[name] = mux(*reg_data, port.addr)
        for name in self._write_ports:
            port = self._port(name)
            enable = None
            if name in self._enable_ports:
                enable = m.bit(self._port(self._enable_ports[name]))
            for i, reg in enumerate(registers):
                mux = m.Mux(2, m.Bits[self._data_width])()
                cond = port.addr == i
                if enable is not None:
                    cond &= enable
                reg_data[i] = mux(reg_data[i], port.data, cond)
            for read_name in self._read_ports:  # forward write
                read_port = self._port(read_name)
                mux = m.Mux(2, m.Bits[self._data_width])()
                cond = port.addr == read_port.addr
                if enable is not None:
                    cond &= enable
                read_data[read_name] = mux(read_data[read_name], port.data,
                                           cond)
        # Commit staged reads and writes.
        for i, reg in enumerate(registers):
            reg.I @= reg_data[i]
        for name in self._read_ports:
            port = self._port(name)
            port.data @= read_data[name]

    def _finalize_verilog(self):
        # TODO: Use inline verilog for safety? Ideally we'd avoid generating
        # verilog at this level, so not a priority (long term fix would be to
        # move this logic to the compiler internals somehow)
        read_data = {}
        for name in self._read_ports:
            self._port(name).data.undriven()
            self._port(name).addr.unused()
            read_data[name] = f"data[{name}_addr]"

        write_port_str = ""
        for name in self._write_ports:
            self._port(name).unused()
            has_enable = name in self._enable_ports
            enable_name = self._enable_ports.get(name, None)
            write_str = f"data[{name}_addr] <= {name}_data;"
            if enable_name:
                write_str = f"""\
if ({enable_name}) begin
    {write_str}
end\
"""
            write_port_str += write_str + "\n"
            for read_name in self._read_ports:
                cond = f"{name}_addr == {read_name}_addr"
                if enable_name:
                    cond += f" & {enable_name}"
                read_data[read_name] = \
                    f"{cond} ? {name}_data : {read_data[read_name]}"
        write_port_str = "\n    ".join(write_port_str.splitlines())

        read_port_str = ""
        for name, data in read_data.items():
            read_port_str += f"assign {name}_data = {data};"

        m.inline_verilog(f"""
reg [{self._data_width - 1}:0] data [{self._height - 1}:0];
always @(posedge CLK) begin
    {write_port_str}
end
{read_port_str}
""")

    @m.builder_method
    def _finalize(self):
        if self.backend == "magma":
            self._finalize_magma()
        elif self.backend == "verilog":
            self._finalize_verilog()
        else:
            raise ValueError(f"Unsupported backend: {self.backend}")



from mantle import DefineEQ, DefineNE
from mantle import DefineULE, DefineUGE, DefineUGT, DefineULT
from mantle import DefineSLE, DefineSGE, DefineSGT, DefineSLT

def EQ(n, **kwargs):
    return DefineEQ(n)(**kwargs)

def NE(n, **kwargs):
    return DefineNE(n)(**kwargs)

def UGE(n, **kwargs):
    return DefineUGE(n)(**kwargs)

def ULE(n, **kwargs):
    return DefineULE(n)(**kwargs)

def ULT(n, **kwargs):
    return DefineULT(n)(**kwargs)

def UGT(n, **kwargs):
    return DefineUGT(n)(**kwargs)

def SGE(n, **kwargs):
    return DefineSGE(n)(**kwargs)

def SLE(n, **kwargs):
    return DefineSLE(n)(**kwargs)

def SLT(n, **kwargs):
    return DefineSLT(n)(**kwargs)

def SGT(n, **kwargs):
    return DefineSGT(n)(**kwargs)




import operator

import magma as m
from magma import *
from magma.compatibility import IntegerTypes
from hwtypes import BitVector
import mantle.primitives
from .logic import DefineFoldOp, get_length, Invert, Not
from .util import DeclareCoreirCircuit, get_int_vector_type
from magma.logging import root_logger


LOGGER = root_logger()


def declare_binop(name, python_op, out_type=None, signed=False):
    def simulate(self, value_store, state_store):
        I0 = get_int_vector_type(signed)(value_store.get_value(self.I0))
        I1 = get_int_vector_type(signed)(value_store.get_value(self.I1))
        O = python_op(I0, I1).as_bool_list()
        if out_type is Bit:
            assert len(O) == 1, "out_type is Bit but the operation returned a list of length {}".format(len(O))
            O = O[0]
        value_store.set_value(self.O, O)

    def Declare(width, T=Bits):
        if width is None:
            T = Bit
            return DeclareCoreirCircuit("coreir_bit_{}{}".format(name, width),
                                  'I0', In(T), 'I1', In(T),
                                  'O', Out(out_type if out_type else T),
                                  stateful=False,
                                  simulate=simulate,
                                  verilog_name="coreir_" + name,
                                  coreir_name=name,
                                  coreir_lib = "corebit")
        else:
            if issubclass(T, m.BFloat):
                coreir_lib = "float"
                if T.N != 16:
                    raise NotImplementedError("Only BFloat16 supported")
                coreir_genargs = {"exp_bits": 8, "frac_bits": 7}
            else:
                coreir_lib = "coreir"
                coreir_genargs = {"width": width}
            T = T[width]
            return DeclareCoreirCircuit("coreir_{}{}".format(name, width),
                                  'I0', In(T), 'I1', In(T),
                                  'O', Out(out_type if out_type else T),
                                  stateful=False,
                                  simulate=simulate,
                                  verilog_name="coreir_" + name,
                                  coreir_name=name,
                                  coreir_lib = coreir_lib,
                                  coreir_genargs=coreir_genargs)

    return Declare

DefineCoreirMul = declare_binop("mul", operator.mul)

def DefineCoreirAdd(width, T=m.Bits):
    def simulate_coreir_add(self, value_store, state_store):
        I0 = BitVector[width](value_store.get_value(self.I0))
        I1 = BitVector[width](value_store.get_value(self.I1))
        value_store.set_value(self.O, I0 + I1)
    if issubclass(T, m.BFloat):
        coreir_lib = "float"
        if T.N != 16:
            raise NotImplementedError("Only BFloat16 supported")
        coreir_genargs = {"exp_bits": 8, "frac_bits": 7} # , "has_cout": has_cout, "has_cin": has_cin}
    else:
        coreir_lib = "coreir"
        coreir_genargs = {"width": width} # , "has_cout": has_cout, "has_cin": has_cin}
    T = T[width]
    coreir_io = ['I0', In(T),
                 'I1', In(T),
                 'O', Out(T)]
    return DeclareCoreirCircuit(f"coreir_add{width}", *coreir_io,
            coreir_name="add", coreir_lib=coreir_lib,
            coreir_genargs=coreir_genargs,
            simulate=simulate_coreir_add)

@cache_definition
def DefineAdd(N=None, cout=False, cin=False, width=None, T=m.Bits):
    if N is None:
        if width is None:
            raise ValueError("Either N or width must be not None")
        N = width
    elif width is not None:
        LOGGER.warning("Both N and width are not None, using N")
    has_cout = cout
    has_cin = cin
    if not has_cout and not has_cin:
        return DefineCoreirAdd(N, T)
    class Add(mantle.primitives.DeclareAdd(N, cin=has_cin, cout=has_cout)):
        @classmethod
        def definition(add):
            width = N
            if has_cout:
                width += 1
            CoreirAdd = DefineCoreirAdd(width, T)
            coreir_add = CoreirAdd()
            I0 = add.I0
            I1 = add.I1
            if has_cout:
                I0 = concat(add.I0, bits(0, n=1))
                I1 = concat(add.I1, bits(0, n=1))
            if has_cin:
                coreir_add_cin = CoreirAdd()
                wire(coreir_add_cin.I0, concat(bits(add.CIN), bits(0,
                    n=width-1)))
                wire(coreir_add_cin.I1, I0)
                I0 = coreir_add_cin.O
            wire(I0, coreir_add.I0)
            wire(I1, coreir_add.I1)
            O = coreir_add.O
            if has_cout:
                COUT = O[-1]
                O = O[:-1]
            wire(O, add.O)
            if has_cout:
                wire(COUT, add.COUT)
    return Add


@cache_definition
def DefineSub(N, cout=False, cin=False, T=m.Bits):
    has_cout = cout
    has_cin = cin
    class Sub(mantle.primitives.DeclareSub(N, cin=has_cin, cout=has_cout, T=T)):
        @classmethod
        def definition(io):
            invert = Invert(N)
            add = DefineAdd(N, cin=True, cout=cout)()
            wire(io.I0, add.I0)
            wire(io.I1, invert.I)
            wire(invert.O, add.I1)
            wire(add.O, io.O)
            if cin:
                wire( Not()(io.CIN), add.CIN )
            else:
                wire( 1, add.CIN )
            if cout:
                wire(add.COUT, io.COUT)
    return Sub


def DefineNegate(width):
    T = Bits[width]

    class _Negate(mantle.primitives.DeclareNegate(width)):
        @classmethod
        def definition(neg):
            args = {"in": In(T), "out": Out(T)}

            class _CoreirNeg(Circuit):
                name = "coreir_neg"
                io = m.IO(**args)
                coreir_name = "neg"
                coreir_lib = "coreir"
                coreir_genargs = {"width": width}

            coreir_neg = _CoreirNeg()
            wire(neg.I, getattr(coreir_neg, "in"))
            wire(neg.O, coreir_neg.out)

    return _Negate


def DefineASR(width):
    T = Bits[width]

    class _ASR(mantle.primitives.DeclareASR(width)):
        @classmethod
        def definition(asr):
            args = {"in0": In(T), "in1": In(T), "out": Out(T)}

            class _CoreirASR(Circuit):
                name = "coreir_" + asr.name
                io = m.IO(**args)
                coreir_name="ashr"
                coreir_lib="coreir"
                coreir_genargs={"width": width}

            coreir_asr = _CoreirASR()
            wire(asr.I0, coreir_asr.in0)
            wire(asr.I1, coreir_asr.in1)
            wire(asr.O, coreir_asr.out)

    return _ASR

def ASR(width, **kwargs):
    return DefineASR(width)(**kwargs)


DefineCoreirMul = declare_binop("mul", operator.mul)

DefineMul = DefineCoreirMul


DefineCoreirUDiv = declare_binop("udiv", operator.truediv)

DefineUDiv = DefineCoreirUDiv


DefineCoreirSDiv = declare_binop("sdiv", operator.truediv)

DefineSDiv = DefineCoreirSDiv


DefineCoreirUMod = declare_binop("urem", operator.mod)

DefineUMod = DefineCoreirUMod


DefineCoreirSMod = declare_binop("srem", operator.mod)

DefineSMod = DefineCoreirSMod



from magma import *
from mantle import And
from .ring import Ring

__all__ = ['CascadedRing']

def CascadedRing(nlist, has_ce=False, **kwargs):
    """
    n cascaded shift registers in a ring
    """

    if has_ce:
        ce = In(Bit)()
        args = ['CE', ce]
    else:
        ce = 1
        args = []

    for i in range(len(nlist)):
        n = nlist[i] - 1
        assert(n < 32)
        ring = Ring(n, has_ce=True)
        ring(CE=ce)
        and2 = And(2)
        and2(ce, ring.O[n-1])
        ce = and2.O

    args += ["O", ce]
    return AnonymousCircuit(args)




import magma as m
from mantle import EQ, LUT


__all__  = ['DefineDecode', 'Decode', 'decode']


@m.cache_definition
def DefineDecode(i, n):

    class _Decode(m.Circuit):
        name = f"Decode{i}{n}"
        io = m.IO(I=m.In(m.Bits[n]), O=m.Out(m.Bit))
        m.wire(io.O, EQ(n)(io.I, m.bits(i, n)))

    return _Decode


def Decode(i, n, invert=False, **kwargs):
    """
    Decode the n-bit number i.

    @return: 1 if the n-bit input equals i
    """

    if n <= 4:
        i = 1 << i
        if invert:
            mask = (1 << (1<<n)) - 1
            i = mask & (~i)
        return m.uncurry(LUT(i, n, **kwargs))

    return DefineDecode(i, n)()


def decode(I, i, invert=False, **kwargs):
    return Decode(i, len(I), invert=invert, **kwargs)(I)



from collections.abc import Sequence
from magma import *
from magma.compatibility import IntegerTypes
from magma.bitutils import lutinit, int2seq
from ..ice40.PLB import SB_LUT4, A0, A1, A2, A3

__all__  = ['LUT1', 'LUT2', 'LUT3', 'LUT4']
__all__ += ['LUT5', 'LUT6', 'LUT7', 'LUT8']
__all__ += ['LUTN', 'LUT']
__all__ += ['A0', 'A1', 'A2', 'A3']


def Mux2(**kwargs):
    """Construct a Mux with 2 1-bit inputs."""
    MUX2DATA = (~A2&A0)|(A2&A1)
    lut = LUT3(MUX2DATA, **kwargs)
    return AnonymousCircuit("I0", lut.I0,
                            "I1", lut.I1,
                            "S", lut.I2,
                            "O", lut.O)

def LUT1(rom, **kwargs):
    lut = SB_LUT4(LUT_INIT=(lutinit(rom,1<<1)[0], 16),**kwargs)
    wire(0, lut.I1)
    wire(0, lut.I2)
    wire(0, lut.I3)
    return AnonymousCircuit("I0", lut.I0, "O", lut.O)

def LUT2(rom, **kwargs):
    lut = SB_LUT4(LUT_INIT=(lutinit(rom,1<<2)[0], 16), **kwargs)
    wire(0, lut.I2)
    wire(0, lut.I3)
    return AnonymousCircuit("I0", lut.I0, "I1", lut.I1, "O", lut.O)

def LUT3(rom, **kwargs):
    lut = SB_LUT4(LUT_INIT=(lutinit(rom,1<<3)[0], 16), **kwargs)
    wire(0, lut.I3)
    return AnonymousCircuit("I0", lut.I0, "I1", lut.I1, "I2", lut.I2, "O", lut.O)

def LUT4(rom, **kwargs):
    return SB_LUT4(LUT_INIT=lutinit(rom,1<<4), **kwargs)

def LUT5(rom, **kwargs):
    if isinstance(rom, IntegerTypes):
        rom = int2seq(rom, 32)
    I0, I1, I2, I3, I4 = In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)()
    lut = fork( LUT4(rom[ 0:16]), LUT4(rom[16:32]) )
    lut(I0, I1, I2, I3)
    mux = Mux2()
    mux(lut.O[0], lut.O[1], I4)
    return AnonymousCircuit("I0", I0,
                            "I1", I1,
                            "I2", I2,
                            "I3", I3,
                            "I4", I4,
                            "O", mux.O)


def LUT6(rom, **kwargs):
    if isinstance(rom, IntegerTypes):
        rom = int2seq(rom, 64)
    I0, I1, I2, I3, I4, I5 = In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)()
    lut = fork( LUT5(rom[ 0:32]), LUT5(rom[32:64]))
    mux = Mux2()
    lut(I0, I1, I2, I3, I4)
    mux(lut.O[0], lut.O[1], I5)
    return AnonymousCircuit("I0", I0,
                            "I1", I1,
                            "I2", I2,
                            "I3", I3,
                            "I4", I4,
                            "I5", I5,
                            "O", mux.O)

def LUT7(rom, **kwargs):
    if isinstance(rom, IntegerTypes):
        rom = int2seq(rom, 128)
    I0, I1, I2, I3, I4, I5, I6 = In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)()
    lut = fork( LUT6(rom[ 0:64]), LUT6(rom[64:128]))
    mux = Mux2()
    lut(I0, I1, I2, I3, I4, I5)
    mux(lut.O[0], lut.O[1], I6)
    return AnonymousCircuit("I0", I0,
                            "I1", I1,
                            "I2", I2,
                            "I3", I3,
                            "I4", I4,
                            "I5", I5,
                            "I6", I6,
                            "O", mux.O)

def LUT8(rom, **kwargs):
    if isinstance(rom, IntegerTypes):
        rom = int2seq(rom, 256)
    I0, I1, I2, I3, I4, I5, I6, I7 = In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)(), In(Bit)()
    lut = fork( LUT7(rom[ 0:128]), LUT7(rom[128:256]))
    mux = Mux2()
    lut(I0, I1, I2, I3, I4, I5, I6)
    mux(lut.O[0], lut.O[1], I7)
    return AnonymousCircuit("I0", I0,
                            "I1", I1,
                            "I2", I2,
                            "I3", I3,
                            "I4", I4,
                            "I5", I5,
                            "I6", I6,
                            "I7", I7,
                            "O", mux.O)

def LUTN(rom, n=None, **kwargs):
    """
    n-bit LUT

    I : In(Bits(n)), O : Out(Bit)
    """

    if n is not None:
        n = 1 << n

    if isinstance(rom, Sequence):
        if n is None:
            n = len(rom)
        else:
            assert n == len(rom)

    assert n is not None


    if n == 2:
        return LUT1(rom, **kwargs)
    if n == 4:
        return LUT2(rom, **kwargs)
    if n == 8:
        return LUT3(rom, **kwargs)
    if n == 16:
        return LUT4(rom, **kwargs)
    if n == 32:
        return LUT5(rom, **kwargs)
    if n == 64:
        return LUT6(rom, **kwargs)
    if n == 128:
        return LUT7(rom, **kwargs)
    if n == 256:
        return LUT8(rom, **kwargs)

    return None

LUT = LUTN





from typing import Optional

import magma as m


def _zero_init_args(T):
    if issubclass(T, m.Tuple):
        return [_zero_init_args(v) for v in T.types()]
    if issubclass(T, m.Array):
        return [_zero_init_args(T.T) for _ in range(len(T))]
    return [0]


class SIPO(m.Generator2):
    def __init__(self, n: int, T: m.Kind = m.Bit, init=None,
                 has_enable: bool = False,
                 reset_type: Optional[m.AbstractReset] = None):
        if init is None:
            init = [T(*_zero_init_args(T)) for _ in range(n)]
        self.name = f"SIPO{n}"
        self.io = m.IO(
            I=m.In(T),
            O=m.Out(m.Array[n, T] if T is not m.Bit else m.Bits[n])
        )

        # TODO: Add magma helper func for this
        has_async_reset = reset_type == m.AsyncReset
        has_async_resetn = reset_type == m.AsyncResetN
        has_reset = reset_type == m.Reset
        has_resetn = reset_type == m.ResetN
        self.io += m.ClockIO(has_enable=has_enable,
                             has_async_reset=has_async_reset,
                             has_async_resetn=has_async_resetn,
                             has_reset=has_reset, has_resetn=has_resetn)

        regs = (m.Register(T, init=init[i], has_enable=has_enable,
                           reset_type=reset_type)() for i in range(n))
        # TODO: Default clock wiring logic raises warning inside scan
        self.io.O @= m.scan(regs, scanargs={"I": "O"})(self.io.I)



import magma as m
from mantle import XOr, And, Or


__all__ = [ "FullAdder" ]


class FullAdder(m.Circuit):
    io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), CIN=m.In(m.Bit),
              O=m.Out(m.Bit), COUT=m.Out(m.Bit))

    # Generate the sum
    m.wire(XOr(3)(io.I0, io.I1, io.CIN), io.O)
    # Generate the carry
    m.wire(Or(3)(And(2)(io.I0, io.I1),
                 And(2)(io.I1, io.CIN),
                 And(2)(io.I0, io.CIN)), io.COUT)



from magma import *
from ..ice40.PLB import SB_CARRY, A0, A1, A2
from .LUT import LUT3

__all__  = ["FullAdder", 'fulladder']

class FullAdder(Circuit):
    IO = ["I0", In(Bit), "I1", In(Bit), "CIN", In(Bit), "O", Out(Bit), "COUT", Out(Bit)]
    @classmethod
    def definition(io):
        sum = LUT3(A0^A1^A2)
        carry = SB_CARRY() # (A0&A1)|(A1&A2)|(A2&A0)
        wire( sum(io.I0,io.I1,io.CIN), io.O )
        wire( carry(io.I0,io.I1,io.CIN), io.COUT )

def fulladder(a, b, c, **kwargs):
    return FullAdder()(a, b, c, **kwargs)



"""
These are imported at the end

This allows targets to override the implementations of generic primitives
(e.g. DefineRAM).
"""
from mantle import DefineRAM, DefineDualRAM

__all__  = ['RAM', 'DualRAM']

def RAM(height, width, **kwargs):
    return DefineRAM(height, width)(**kwargs)

def DualRAM(height, width, **kwargs):
    return DefineDualRAM(height, width)(**kwargs)



import magma as m
from mantle import XOr, And


__all__ = [ "HalfAdder" ]


class HalfAdder(m.Circuit):
    io = m.IO(I=m.In(m.Bit), CIN=m.In(m.Bit), O=m.Out(m.Bit), COUT=m.Out(m.Bit))

    # Generate the sum
    m.wire( XOr(2)(io.I, io.CIN), io.O )
    # Generate the carry
    m.wire( And(2)(io.I, io.CIN), io.COUT )



from __future__ import division


from magma import *
import magma as m
from magma.backend.coreir_ import CoreIRBackend
from magma.frontend.coreir_ import DefineCircuitFromGeneratorWrapper, GetCoreIRBackend
from .util import DeclareCoreirCircuit
from hwtypes import BitVector

import math


def DefineCoreirMux(width=None):
    N = width
    def simulate(self, value_store, state_store):
        I0 = BitVector[N](value_store.get_value(self.I0))
        I1 = BitVector[N](value_store.get_value(self.I1))
        S = BitVector[N](value_store.get_value(self.S))
        O = I1 if S.as_int() else in0
        value_store.set_value(self.O, O)
    if width is None:
        return DeclareCoreirCircuit("coreir_mux{}".format(N),
            *["I0", In(Bit), "I1", In(Bit), "S", In(Bit),
             "O", Out(Bit)],
            verilog_name="coreir_bitmux",
            coreir_name="mux",
            coreir_lib="corebit",
            simulate=simulate
        )
    else:
        return DeclareCoreirCircuit("coreir_mux{}".format(N),
            *["I0", In(Bits[N]), "I1", In(Bits[N]), "S", In(Bit),
             "O", Out(Bits[N])],
            verilog_name="coreir_mux",
            coreir_name="mux",
            coreir_lib="coreir",
            simulate=simulate,
            coreir_genargs={"width": N}
        )


is_power_of_two = lambda num: num != 0 and ((num & (num - 1)) == 0)


@m.cache_definition
def _declare_muxn(height, width):
    def simulate(self, value_store, state_store):
        sel = BitVector[m.bitutils.clog2(height)](value_store.get_value(self.I.sel))
        out = BitVector[width](value_store.get_value(self.I.data[int(sel)]))
        value_store.set_value(self.O, out)
    return DeclareCoreirCircuit(f"coreir_commonlib_mux{height}x{width}",
        *["I", In(Product.from_fields("anon",
                                      dict(data=Array[height, Bits[width]],
                                           sel=Bits[m.bitutils.clog2(height)]))),
          "O", Out(Bits[width])],
        coreir_name="muxn",
        coreir_lib="commonlib",
        simulate=simulate,
        coreir_genargs={"width": width, "N": height}
    )


@m.cache_definition
def DefineMux(height=2, width=None, T=None):
    if T is not None:
        assert width is None, "Can only specify width **or** T"
        # Sanitize names for verilog by removing parens
        # TODO: Make this a reuseable feature
        suffix = str(T).replace("(", "").replace(")", "").replace(",", "_").replace("=", "_").replace("[", "").replace("]", "").replace(" ", "")
        T = T
    else:
        suffix = f"{width}"
        if width is None:
            T = Bit
        else:
            T = Bits[width]

    decl = []
    for i in range(height):
        decl += ["I{}".format(i), In(T)]
    if height == 2:
        select_type = Bit
    else:
        select_type = Bits[m.bitutils.clog2(height)]
    decl += ['S', In(select_type)]
    decl += ['O', Out(T)]

    class _Mux(Circuit):
        name = "Mux{}x{}".format(height, suffix)
        io = m.IO(**dict(zip(decl[::2], decl[1::2])))

        # NOTE(rsetaluri): Because of scoping rules in the class definition, we
        # can not use 'io' in list comprehensions. Therfore we use simple
        # iteration-based list construction.
        if T is not None and not (issubclass(T, m.Digital) or issubclass(T, m.Array) and issubclass(T.T, m.Bit)):
            if issubclass(T, m.Tuple):
                for i in range(len(T.keys())):
                    Is = []
                    for j in range(height):
                        Is.append(getattr(io, f"I{j}")[list(T.keys())[i]])
                    io.O[i] <= DefineMux(height, T=list(T.types())[i])()(*Is, io.S)
            else:
                assert issubclass(T, m.Array), f"Expected array or type type, got {T}, type is {type(T)}"
                for i in range(len(T)):
                    Is = []
                    for j in range(height):
                        Is.append(getattr(io, f"I{j}")[i])
                    io.O[i] <= DefineMux(height, T=type(Is[0]))()(*Is, io.S)
        else:
            if T is None and width is None or issubclass(T, m.Digital):
                mux = _declare_muxn(height, 1)()
            else:
                mux = _declare_muxn(height, width if T is None else len(T))()
            for i in range(height):
                if T is None and width is None or issubclass(T, m.Digital):
                    m.wire(getattr(io, f"I{i}"), mux.I.data[i][0])
                else:
                    m.wire(getattr(io, f"I{i}"), mux.I.data[i])
            if height == 2:
                m.wire(io.S, mux.I.sel[0])
            else:
                m.wire(io.S, mux.I.sel)
            if T is None and width is None or issubclass(T, m.Digital):
                wire(mux.O[0], io.O)
            else:
                wire(mux.O, io.O)
    return _Mux


def Mux(height=2, width=None, T=None, **kwargs):
    return DefineMux(height, width, T=T)(**kwargs)

Mux2 = DefineMux(2)
Mux4 = DefineMux(4)
Mux8 = DefineMux(8)
Mux16 = DefineMux(16)

@cache_definition
def DefineCommonlibMuxN(N: int, width: int):
    """
    Get a Mux that handles any number of inputs

    :param N: The number of inputs to select between
    :param width: The number of bits per input
    :return: A circuit with the following ports:
    I : {
        data: In(Array(N, Array(width, Bit)))
        sel: In(Array(getRAMAddrWidth(N), Bit))
    }
    out : Out(Array(width, Bit))
    Note: even though this isn't a RAM, the AddrWidth computation is the same.
    """
    name = "CommonlibMuxN_n{}_w{}".format(str(N), str(width))
    return DefineCircuitFromGeneratorWrapper(GetCoreIRBackend(), "commonlib", "muxn",
                                             name, ["mantle", "coreir", "global"],
                                             {"N": N, "width": width})

def CommonlibMuxN(N: int, width: int):
    return DefineCommonlibMuxN(N, width)()



import os
from magma import *
from mantle import *

__all__ = ['LHCA']

_lhcataps = {}


def LHCA(n, init=1, has_ce=False):

    def readtaps():
        global _lhcataps

        if len(_lhcataps) != 0:
            return

        import csv

        name = os.path.join(os.path.dirname(__file__), 'lhca.csv')

        f = open(name)

        row = csv.reader(f)

        for data in row:
            if data[0].isdigit():
                tap = int(data[0])
                _lhcataps[tap] = [int(v) for v in data[1].split(',')]

    readtaps()

    taps = _lhcataps[n]

    reg = Register(n, init=init, has_ce=has_ce)

    def lut(y):
        tap = (y + 1) in taps
        if y == 0:
            if tap:
                return LUT2(A0^A1)(reg.O[y], reg.O[y + 1])
            else:
                return LUT1(A0)(reg.O[y + 1])
        elif y == n - 1:
            if tap:
                return LUT2(A0^A1)(reg.O[y - 1], reg.O[y])
            else:
                return LUT1(A0)(reg.O[y - 1])
        else:
            if tap:
                return LUT3(A0^A1^A2)(reg.O[y - 1], reg.O[y], reg.O[y + 1])
            else:
                return LUT2(A0^A1)(reg.O[y - 1], reg.O[y + 1])

    # col returns a list of outputs, convert to an array
    luts = array(col(lut, n))
    reg(luts)

    args = ["O", reg.O] + reg.interface.clockargs()

    return AnonymousCircuit(args)
